import json
import time
import urllib
from typing import Any, Dict

import pandas as pd
import requests
from prefect.utilities import logging

from ..config import local_config
from .base import Source

logger = logging.get_logger(__name__)

API_ENDPOINT = "https://api.supermetrics.com/enterprise/v2/query/data/json"
DEFAULT_CREDENTIALS = local_config.get("SUPERMETRICS")


class Supermetrics(Source):
    """
    A class implementing the Supermetrics API.

    Documentation for this API is located at: https://supermetrics.com/docs/product-api-getting-started/

    Parameters
    ----------
    query_params : Dict[str, Any], optional
        The parameters to pass to the GET query.
        See https://supermetrics.com/docs/product-api-get-data/ for full specification,
        by default None
    """

    def __init__(self, *args, query_params: Dict[str, Any] = None, **kwargs):
        credentials = kwargs.pop("credentials", DEFAULT_CREDENTIALS)
        super().__init__(*args, credentials=credentials, **kwargs)
        self.query_params = query_params

    @classmethod
    def get_params_from_api_query(cls, url: str) -> Dict[str, Any]:
        url_unquoted = urllib.parse.unquote(url)
        s = urllib.parse.parse_qs(url_unquoted)
        params = s[
            "https://api.supermetrics.com/enterprise/v2/query/data/powerbi?json"
        ][0]
        params_d = json.loads(params)
        return params_d

    @classmethod
    def from_url(cls, url: str, credentials: Dict[str, Any] = None):
        obj = Supermetrics(credentials=credentials)
        params = cls.get_params_from_api_query(url)
        obj.query_params = params
        obj.query_params["api_key"] = credentials["API_KEY"]
        return obj

    def to_json(self) -> Dict[str, Any]:

        if not self.query_params:
            raise ValueError("Please build the query first")

        self.query_params["api_key"] = self.credentials["API_KEY"]
        params = {"json": json.dumps(self.query_params)}
        try:
            response = requests.get(API_ENDPOINT, params=params)
        except ConnectionError as e:
            # retry once
            logger.warning(f"Error: \n\n{e} \n\nRetrying once in 5s...")
            time.sleep(5)
            response = requests.get(API_ENDPOINT, params=params)
        assert response.ok

        response_json = response.json()

        return response_json

    def to_df(self) -> pd.DataFrame:
        data = self.to_json()["data"]
        df = pd.DataFrame(data[1:], columns=data[0])
        return df

    def query(self, params: Dict[str, Any]):
        self.query_params = params
        self.query_params["api_key"] = self.credentials["API_KEY"]
        return self



import prefect
from prefect import Task
from ..sources import AzureBlobStorage


class BlobFromCSV(Task):
    def __init__(self, *args, **kwargs):
        super().__init__(name="csv_to_blob_storage", *args, **kwargs)

    def __call__(self):
        """Generate a blob from a local CSV file"""

    def run(self, from_path: str, to_path: str):

        logger = prefect.context.get("logger")

        blob_storage = AzureBlobStorage()

        logger.info(f"Copying from {from_path} to {to_path}...")
        blob_storage.to_storage(from_path=from_path, to_path=to_path)
        logger.info(f"Successfully uploaded data to {to_path}.")



import prefect
from prefect import Task
from typing import Dict, Any, Literal
from ..sources import AzureSQL


class CreateTableFromBlob(Task):
    def __init__(self, *args, **kwargs):
        super().__init__(name="blob_to_azure_sql", *args, **kwargs)

    def __call__(self):
        """Bulk insert a CSV into an Azure SQL table"""

    def run(
        self,
        blob_path: str,
        schema: str,
        table: str,
        dtypes: Dict[str, Any],
        if_exists: Literal = ["fail", "replace"],
    ):
        """Create a table from an Azure Blob object.
        Currently, only CSV files are supported.

        Parameters
        ----------
        blob_path : str
            Path to the blob, eg. 'container_name/path/to.csv'.
        schema : str
            Destination schema.
        table : str
            Destination table.
        dtypes : Dict[str, Any]
            Data types to force.
        if_exists : Literal, optional
            What to do if the table already exists, by default ["fail", "replace"]
        """
        logger = prefect.context.get("logger")

        # create table
        azure_sql = AzureSQL()
        azure_sql.create_table(
            schema=schema, table=table, dtypes=dtypes, if_exists=if_exists
        )
        fqn = f"{schema}.{table}" if schema else table
        logger.info(f"Successfully created table {fqn}.")

        # insert data
        azure_sql.bulk_insert(
            schema=schema,
            table=table,
            source_path=blob_path,
            if_exists=if_exists,
        )
        logger.info(f"Successfully inserted data into {fqn}.")



from typing import Any, Dict, List, Literal

import pyodbc
from prefect.utilities import logging

from ..config import local_config
from .base import Source

logger = logging.get_logger(__name__)

DEFAULT_CREDENTIALS = local_config.get("AZURE_SQL")
DEFAULT_SCHEMA = "sandbox"


class AzureSQL(Source):
    """
    Azure SQL Database connector (do not confuse with managed SQL Server
    that's also provided as a service by Azure).

    Parameters
    ----------
    server : str, optional
        Server address, eg. mydb.database.windows.net, by default None
    db : str, optional
        Name of the database, by default None
    user : str, optional
        Username, by default None
    pw : str, optional
        Password, by default None
    """

    def __init__(
        self,
        server: str = None,
        db: str = None,
        user: str = None,
        pw: str = None,
        *args,
        **kwargs,
    ):
        credentials = kwargs.pop("credentials", DEFAULT_CREDENTIALS)
        super().__init__(*args, credentials=credentials, **kwargs)

        # Connection details.
        self.driver = "ODBC Driver 17 for SQL Server"
        server = server or self.credentials["server"]
        db = db or self.credentials["db_name"]
        user = user or self.credentials["user"]
        pw = pw or self.credentials["password"]

        self.conn_str = f"DRIVER={{{self.driver}}};SERVER={{{server}}};DATABASE={{{db}}};UID={{{user}}};PWD={{{pw}}}"
        self._con = None

    @property
    def con(self):
        """A singleton for connection"""
        if not self._con:
            self._con = pyodbc.connect(self.conn_str)
        return self._con

    def run(self, query: str):
        cursor = self.con.cursor()
        cursor.execute(query)
        if query.upper().startswith("SELECT"):
            return cursor.fetchall()
        self.con.commit()

    @property
    def schemas(self) -> List[str]:
        schemas_tuples = self.run("SELECT s.name as schema_name from sys.schemas s")
        return [schema_tuple[0] for schema_tuple in schemas_tuples]

    @property
    def tables(self) -> List[str]:
        tables_tuples = self.run("SELECT * FROM information_schema.tables")
        return [table for row in tables_tuples for table in row]

    def create_table(
        self,
        table: str,
        schema: str = DEFAULT_SCHEMA,
        dtypes: Dict[str, Any] = None,
        if_exists: Literal = ["fail", "replace"],
    ):
        fqn = f"{schema}.{table}"
        indent = "  "
        dtypes_rows = [indent + col + " " + dtype for col, dtype in dtypes.items()]
        dtypes_formatted = ",\n".join(dtypes_rows)
        create_table_sql = f"CREATE TABLE {fqn}(\n{dtypes_formatted}\n)"
        if if_exists == "replace":
            try:
                self.run(f"DROP TABLE {schema}.{table}")
            except:
                pass
        self.run(create_table_sql)
        return True

    def bulk_insert(
        self,
        table: str,
        schema: str = DEFAULT_SCHEMA,
        source_path: str = None,
        if_exists: Literal = "append",
    ):
        fqn = f"{schema}.{table}"
        insert_sql = f"""
            BULK INSERT {fqn} FROM '{source_path}'
            WITH (
                CHECK_CONSTRAINTS,
                DATA_SOURCE = '{self.credentials['data_source']}',
                DATAFILETYPE='char',
                FIELDTERMINATOR=',',
                ROWTERMINATOR='0x0a',
                FIRSTROW=2,
                KEEPIDENTITY,
                TABLOCK
            );
        """
        if if_exists == "replace":
            self.run(f"DELETE FROM {schema}.{table}")
        self.run(insert_sql)
        return True

    def create_external_database(url):
        """TODO"""

        credential_name = "data_lake_credential"
        shared_access_token = ""

        # "CREATE MASTER KEY ENCRYPTION BY PASSWORD = <enter very strong password here>""

        create_external_db_credential_sql = f"""
        USE [acdb]
        CREATE DATABASE SCOPED CREDENTIAL {credential_name}
        WITH IDENTITY = 'SHARED ACCESS SIGNATURE'
        SECRET = '{shared_access_token}';
        """

        create_external_db_sql = f"""
        USE [acdb]
        CREATE EXTERNAL DATA SOURCE testing WITH (
        LOCATION = 'https://dyvenia1.blob.core.windows.net/tests',
        CREDENTIAL = {credential_name}
        );
        """

        pass



import prefect
from prefect import task, Flow
from prefect.storage import GitHub
from prefect.run_configs import DockerRun


@task
def say_hello():
    logger = prefect.context.get("logger")
    logger.info("Hello!")


@task
def say_bye():
    logger = prefect.context.get("logger")
    logger.info("Bye!")


STORAGE = GitHub(
    repo="dyvenia/fiotto",
    path="fiotto/examples/hello_world.py",
    access_token_secret="github_token",  # name of the Prefect secret with the GitHub token
)
RUN_CONFIG = DockerRun(
    image="prefecthq/prefect",
    env={"SOME_VAR": "value"},
    labels=["testing"],
)

with Flow("Hello, world!", storage=STORAGE, run_config=RUN_CONFIG) as flow:
    hello = say_hello()
    bye = say_bye()
    bye.set_upstream(hello, flow=flow)


if __name__ == "__main__":
    flow.run()



import prefect
import datetime
from prefect import Task
from viadot.sources import UKCarbonIntensity


class StatsToCSV(Task):
    """A Prefect task for downloading UK Carbon Instensity Statistics (stats) to a csv file."""

    def __init__(self, *args, **kwargs):
        """Generate the task."""
        super().__init__(name="uk_carbon_intensity_stats_to_csv", *args, **kwargs)

    def __call__(self):
        """
        Run the task.

        Parameters
        ----------
        path : str
            Path of the csv file created or edited by this task
        days_back : int, optional
            How many days of stats to download in the csv.
            UK Carbon Intensity statistics are available for up to 30 days,
            by default 10
        """

    def run(self, path: str, days_back: int = 10):
        """
        Run the task.

        Parameters
        ----------
        path : str
            Path of the csv file created or edited by this task
        days_back : int, optional
            How many days of stats to download in the csv.
            UK Carbon Intensity statistics are available for up to 30 days,
            by default 10
        """

        logger = prefect.context.get("logger")
        carbon = UKCarbonIntensity()
        now = datetime.datetime.now()
        logger.info(f"Downloading data to {path}...")
        for i in range(days_back):
            from_delta = datetime.timedelta(days=i + 1)
            to_delta = datetime.timedelta(days=i)
            to = now - to_delta
            from_ = now - from_delta
            carbon.query(f"/intensity/stats/{from_.isoformat()}/{to.isoformat()}")
            carbon.to_csv(path, if_exists="append")

        # Download data to a local CSV file
        logger.info(f"Successfully downloaded data to {path}.")



from prefect import Flow
from ..tasks import SupermetricsToCSV, BlobFromCSV, CreateTableFromBlob
from typing import Dict, List, Any

supermetrics_to_csv_task = SupermetricsToCSV()
csv_to_blob_storage_task = BlobFromCSV()
blob_to_azure_sql_task = CreateTableFromBlob()


class SupermetricsToAzureSQL(Flow):
    def __init__(
        self,
        name: str,
        query: Dict[str, Any] = None,
        dtypes: Dict[str, Any] = None,
        blob_path: str = None,
        table: str = None,
        schema: str = None,
        local_file_path: str = None,
        if_exists: str = "replace",
        *args: List[any],
        **kwargs: Dict[str, Any]
    ):
        self.query = query
        self.dtypes = dtypes
        self.blob_path = blob_path
        self.table = table
        self.schema = schema
        self.local_file_path = local_file_path or "test.csv"
        self.if_exists = if_exists
        self.tasks = [
            supermetrics_to_csv_task,
            csv_to_blob_storage_task,
            blob_to_azure_sql_task,
        ]
        super().__init__(*args, name=name, **kwargs)
        self.gen_flow()

    def gen_flow(self):
        supermetrics_to_csv_task.bind(
            query=self.query, path=self.local_file_path, flow=self
        )
        csv_to_blob_storage_task.bind(
            from_path=self.local_file_path, to_path=self.blob_path, flow=self
        )
        blob_to_azure_sql_task.bind(
            blob_path=self.blob_path,
            schema=self.schema,
            table=self.table,
            dtypes=self.dtypes,
            if_exists=self.if_exists,
            flow=self,
        )

        csv_to_blob_storage_task.set_upstream(supermetrics_to_csv_task, flow=self)
        blob_to_azure_sql_task.set_upstream(csv_to_blob_storage_task, flow=self)



from typing import Any, Dict

from azure.storage.blob import BlobServiceClient, ContentSettings

from ..config import local_config
from .base import Source

DEFAULT_CREDENTIALS = local_config.get("AZURE_BLOB_STORAGE")


class AzureBlobStorage(Source):
    """
    A class for pulling data from the Azure Blob Storage
    (do not confuse with Azure Data Lake Gen1 and Azure Data Lake Gen2).

    Documentation for this API is located at: https://supermetrics.com/docs/product-api-getting-started/

    Parameters
    ----------
    credentials : Dict[str, Any], optional
        Credentials containing a connection string, by default None
    """

    def __init__(self, credentials: Dict[str, Any] = None, *args, **kwargs):
        credentials = credentials or DEFAULT_CREDENTIALS
        super().__init__(*args, credentials=credentials, **kwargs)

    def to_storage(self, from_path: str, to_path: str):

        conn_str = self.credentials["CONNECTION_STRING"]
        blob_service_client = BlobServiceClient.from_connection_string(conn_str)

        container_name = to_path.split("/")[0]
        blob_path = "/".join(to_path.split("/")[1:])
        blob_client = blob_service_client.get_blob_client(
            container=container_name, blob=blob_path
        )

        with open(from_path, "rb") as data:
            content_settings = ContentSettings(content_type="text/plain")
            blob_client.upload_blob(data, overwrite=True)
            blob_client.set_http_headers(content_settings=content_settings)

        return True



import json
from os.path import expanduser, join

USER_HOME = expanduser("~")


class Config(dict):
    @classmethod
    def from_json(cls, path: str, key: str = None):
        with open(path) as f:
            config = json.load(f)
            if key:
                config = config[key]
            return cls(**config)


local_config = Config.from_json(join(USER_HOME, ".config", "credentials.json"))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['viadot',
 'viadot.examples',
 'viadot.flows',
 'viadot.sources',
 'viadot.tasks',
 'viadot.tasks.open_apis']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.9,<2.0.0',
 'azure-storage-blob>=12.8.0,<13.0.0',
 'black>=20.8b1,<21.0',
 'install>=1.3.4,<2.0.0',
 'mkdocs-material>=7.1.3,<8.0.0',
 'mkdocs>=1.1.2,<2.0.0',
 'mkdocstrings>=0.15.0,<0.16.0',
 'pandas>=1.2.4,<2.0.0',
 'prefect>=0.14.16,<0.15.0',
 'pyarrow>=3.0.0,<4.0.0',
 'pyodbc>=4.0.30,<5.0.0',
 'requests>=2.25.1,<3.0.0']

setup_kwargs = {
    'name': 'viadot',
    'version': '0.1.2',
    'description': '',
    'long_description': None,
    'author': 'Alessio Civitillo',
    'author_email': 'acivitillo@dyvenia.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from ..config import local_config
from ..flows.supermetrics_to_azure_sql import SupermetricsToAzureSQL
from prefect.storage import github

SUPERMETRICS_CREDENTIALS = local_config.get("SUPERMETRICS")

google_ads_flow = SupermetricsToAzureSQL(
    "Google Ads extract",
    query={
        "ds_id": "AW",
        "ds_accounts": SUPERMETRICS_CREDENTIALS["SOURCES"]["Google Ads"]["Accounts"],
        "ds_user": SUPERMETRICS_CREDENTIALS["USER"],
        "date_range_type": "last_year_inc",
        "fields": [
            "Date",
            "profile",
            "Campaignname",
            "Impressions",
            "Clicks",
            "Cost_eur",
            "SearchImpressionShare",
        ],
        "max_rows": 1000000,
    },
    dtypes={
        "Date": "DATE",
        "Account": "VARCHAR(255)",
        "Campaign": "VARCHAR(255)",
        "Impressions": "FLOAT(24)",
        "Clicks": "FLOAT(24)",
        "Cost": "FLOAT(24)",
        "ImpressionShare": "VARCHAR(255)",
    },
    schema="sandbox",
    table="google_ads",
    blob_path="tests/supermetrics/google_ads.csv",
)

# google_ads_flow.visualize()
google_ads_flow.run()


import prefect
from prefect import Task
from typing import Dict, Any
from ..sources import Supermetrics


class SupermetricsToCSV(Task):
    def __init__(self, *args, **kwargs):
        super().__init__(name="supermetrics_to_csv", *args, **kwargs)

    def __call__(self):
        """Download Supermetrics data to a CSV"""

    def run(self, query: Dict[str, Any], path: str):

        logger = prefect.context.get("logger")

        # Build the URL
        supermetrics = Supermetrics()
        supermetrics.query(query)

        # Download data to a local CSV file
        logger.info(f"Downloading data to {path}...")
        supermetrics.to_csv(path)
        logger.info(f"Successfully downloaded data to {path}.")



import logging
import os
import pandas
from abc import abstractmethod
from typing import Any, Dict

import pyarrow as pa


class Source:
    def __init__(self, *args, credentials: Dict[str, Any] = None, **kwargs):
        self.credentials = credentials
        self.data: pa.Table = None
        self.logger = logging.getLogger(__name__)

    @abstractmethod
    def to_json(self):
        pass

    @abstractmethod
    def to_df():
        pass

    @abstractmethod
    def query():
        pass

    def to_arrow(self):
        df = self.to_df()
        table = pa.Table.from_pandas(df)
        return table

    def to_csv(self, path: str, if_exists="replace"):
        df = self.to_df()
        if if_exists == "append":
            if os.path.isfile(path):
                csv_df = pandas.read_csv(path)
                out_df = pandas.concat([csv_df, df])
            else:
                out_df = df
        elif if_exists == "replace":
            out_df = df
        out_df.to_csv(path, index=False)
        return True
